#include "types.h"
#include "gasdev.h"
#include "rng.h"
#include "statistics.h"
#include <assert.h>
#include "geometry.h"
#include "diffusion.h"

extern long seed;

void getNeighbourCoord(int i, int x, int y, int z, iNeighbours *neigh, eBoundary b, TLattice *lattice, Dimension N[]){

  switch(b){
  case SS:/*cyclic boundaries*/
    if( x == 0 ){ /*LEFT*/
      N[0].x = lattice->xdim-1;      N[0].y = y;      N[0].z = z;
    } else {
      N[0].x = x-1;      N[0].y = y;      N[0].z = z;
    }
    if( y == 0){ /*TOP*/
      N[1].x = x; N[1].y = lattice->ydim-1; N[1].z = z;
    } else {
      N[1].x = x; N[1].y = y-1; N[1].z = z;
    }
    if( x == lattice->xdim-1 ){ /*RIGHT*/
      N[2].x = 0;      N[2].y = y;      N[2].z = z;
    } else {
      N[2].x = x+1;      N[2].y = y;      N[2].z = z;
    }
    if( y == lattice->ydim-1){ /*BOTTOM*/
      N[3].x = x; N[3].y = 0; N[3].z = z;
    } else {
      N[3].x = x; N[3].y = y+1; N[3].z = z;
    }
    if( z == 0 ){ /*FRONT*/
      N[4].x = x;      N[4].y = y;      N[4].z = lattice->zdim-1;
    } else {
      N[4].x = x;      N[4].y = y;      N[4].z = z-1;
    }
    if( z == lattice->zdim-1){ /*BACK*/
      N[5].x = x; N[5].y = y; N[5].z = 0;
    } else {
      N[5].x = x; N[5].y = y; N[5].z = z+1;
    }

    break;
  case R1:/*reflectant*/
    /*    neigh->left = (x==0)?x: x-1;
    neigh->top = (y==0)?y: y-1;
    neigh->right = (x==lattice->xdim-1)?x: x+1;
    neigh->bottom = (y==lattice->ydim-1)?y:y+1;
    neigh->front = (z==0)? z: z-1;
    neigh->back = (z == lattice->zdim-1)? z: z+1;
    break;*/
  case R2:/*reflectant*/
    /*    neigh->left = (x==0)?x: x-1;
    neigh->top = (y==0)?y: y-1;
    neigh->right = (x==lattice->xdim-1)?x: x+1;
    neigh->bottom = (y==lattice->ydim-1)?y:y+1;
    neigh->front = (z==0)? z: z-1;
    neigh->back = (z == lattice->zdim-1)? z: z+1;
    */
    if( x == 0 ){ /*LEFT*/
      N[0].x = x;      N[0].y = y;      N[0].z = z;
    } else {
      N[0].x = x-1;      N[0].y = y;      N[0].z = z;
    }
    if( y == 0){ /*TOP*/
      N[1].x = x; N[1].y = y; N[1].z = z;
    } else {
      N[1].x = x; N[1].y = y-1; N[1].z = z;
    }
    if( x == lattice->xdim-1 ){ /*RIGHT*/
      N[2].x = x;      N[2].y = y;      N[2].z = z;
    } else {
      N[2].x = x+1;      N[2].y = y;      N[2].z = z;
    }
    if( y == lattice->ydim-1){ /*BOTTOM*/
      N[3].x = x; N[3].y = y; N[3].z = z;
    } else {
      N[3].x = x; N[3].y = y+1; N[3].z = z;
    }
    if( z == 0 ){ /*FRONT*/
      N[4].x = x;      N[4].y = y;      N[4].z = z;
    } else {
      N[4].x = x;      N[4].y = y;      N[4].z = z-1;
    }
    if( z == lattice->zdim-1){ /*BACK*/
      N[5].x = x; N[5].y = y; N[5].z = z;
    } else {
      N[5].x = x; N[5].y = y; N[5].z = z+1;
    }
    break;
  case GFILE: 
  case QUARTER:
    /* membrane is reflective. If quarter domain, wall are reflective as well*/
    if( x == 0 ){ /*LEFT*/
	N[0].x = x;      N[0].y = y;      N[0].z = z;
    } else {
      if( lattice->id[x-1][y][z] == EXTERIOR ){
	N[0].x = x;      N[0].y = y;      N[0].z = z;
      } else { 
	N[0].x = x-1;      N[0].y = y;      N[0].z = z;
      }
    }
    
    if( y == 0){/*TOP*/
	N[1].x = x;      N[1].y = y;      N[1].z = z;
    } else {      
      if( lattice->id[x][y-1][z] == EXTERIOR ){
	N[1].x = x;      N[1].y = y;      N[1].z = z;
      } else {
	N[1].x = x;      N[1].y = y-1;      N[1].z = z;
      }
    }
    
    if( x == lattice->xdim - 1 ){ /*RIGHT*/
      N[2].x = x;      N[2].y = y;      N[2].z = z;
    } else {
      if(lattice->id[x+1][y][z] == EXTERIOR ){ 
	N[2].x = x;      N[2].y = y;      N[2].z = z;
      } else {
	N[2].x = x+1;      N[2].y = y;      N[2].z = z;
      }
    }

    if (y == lattice->ydim - 1){/*Bottom*/
      N[3].x = x; N[3].y = y; N[3].z = z;
    } else {
      if(lattice->id[x][y+1][z] == EXTERIOR ){
	N[3].x = x; N[3].y = y; N[3].z = z;
      } else {
	N[3].x = x; N[3].y = y+1; N[3].z = z;
      }
    }

    if(z==0){/*FRONT*/
      N[4].x = x;      N[4].y = y;      N[4].z = z;
    } else {
      if( lattice->id[x][y][z-1] == EXTERIOR ){
	N[4].x = x;      N[4].y = y;      N[4].z = z;
      } else {
	N[4].x = x;      N[4].y = y;      N[4].z = z-1;
      }
    }

    if (z == lattice->zdim -1){/*BACK*/
      N[5].x = x; N[5].y = y; N[5].z = z;
    } else {
      if(lattice->id[x][y][z+1] == EXTERIOR ){
      N[5].x = x; N[5].y = y; N[5].z = z;
      } else {
	N[5].x = x; N[5].y = y; N[5].z = z+1;
      }
    }
    break;
  }
}

void get_membrane_neigh(TLattice *lattice, int x, int y, int z, Dimension *N[9]){

  *N = lattice->N[x][y][z];
  /*
    T->x = lattice->N[x][y][z][TOP].x;
    T->y = lattice->N[x][y][z][TOP].y;
    T->z = lattice->N[x][y][z][TOP].z;
    
    R->x = lattice->N[x][y][z][RIGHT].x;
    R->y = lattice->N[x][y][z][RIGHT].y;
    R->z = lattice->N[x][y][z][RIGHT].z;
    
    B->x = lattice->N[x][y][z][BOTTOM].x;
    B->y = lattice->N[x][y][z][BOTTOM].y;
    B->z = lattice->N[x][y][z][BOTTOM].z;
    
    L->x = lattice->N[x][y][z][LEFT].x;
    L->y = lattice->N[x][y][z][LEFT].y;
    L->z = lattice->N[x][y][z][LEFT].z;
  */
}

void diffusion_membrane_small(int max_l, TDiffusion *diff, TLattice *lattice, int x, int y, int z, int iSpecies){
  double r;
  int l, i;
  double p[4] = {0.25, 0.5, 0.75,1}; /*isotropic membrane diffusion*/
  /*  double p[10] = {0.1111, 0.2222, 0.3333, 0.4444, 0.5555, 0.6666, 0.7777, 0.8888, 0.9999, 1.0};*/
  Dimension *MN[9];

  get_membrane_neigh(lattice, x, y, z, MN);

  for(l=0; l < max_l; l++){
    r = rand3(&seed);
    for(i=0; i < 4; i++){
      if( r < p[i] ){
	lattice->TO[(*MN)[i].x][(*MN)[i].y][(*MN)[i].z].count[iSpecies]++;
	break;
      }
    }
  }
}

void diffusion(TDiffusion *diff, int x, int y, int z,\
	       TLattice *lattice,\
	       TGillespie *gillespie,\
	       double t_alarm[][2],\
	       double t){

  int max_l;
  int mobile_particles;

  int iSpecies;

  for(iSpecies=0; iSpecies < gillespie->N; iSpecies++){
    max_l =lattice->FROM[x][y][z].count[iSpecies];
 
    if( t_alarm[iSpecies][0] == t && max_l > 0){
      if( gillespie->boundedSpecies[iSpecies] == BOUNDED){
	/*do membrane diffusion*/
	  if( max_l < 40){
	    diffusion_membrane_small(max_l, diff, lattice, x, y, z, iSpecies);
	  } else {
	    /*	    diffusion_membrane_large(lattice, x, y, z, diff, iSpecies);*/
	    diffusion_membrane_small(max_l, diff, lattice, x, y, z, iSpecies);
	  } 
      } else {
	/* do cytosol diffusion */
	if( max_l < 60){
	  diffusion_few(max_l, diff, x, y, z, lattice, iSpecies);
	} else {
	  mobile_particles = max_l - rd_round((double)gasdev(&seed,\
							     max_l*diff->p[0],\
							     max_l*diff->p[0] * (1-diff->p[0])));
	  if( mobile_particles < 60){
	    diffusion_few(max_l, diff, x,y,z, lattice, iSpecies);
	  } else{
	    diffusion_large(x, y, z, lattice, diff, iSpecies, mobile_particles);
	  }
	}

      }
    } else {
      /*just copy the count number for the iSpecies*/
      lattice->TO[x][y][z].count[iSpecies] += lattice->FROM[x][y][z].count[iSpecies];
    }
  } /*end for*/
  
}


void diffusion_large(int x, int y, int z,\
		     TLattice *lattice,\
		     TDiffusion *diff,\
		     int iSpecies,
		     int mobile_particles){
  int l, n0, n_m;
  int g1, g2, g3, gg1, gg2, gg3, gg12, gg22, gg32; 
  iNeighbours neigh;
  tSite (*TO)[L_MAX_Y_SIZE][L_MAX_Z_SIZE];
  double *p;
  Dimension N[6]; /*neighbours coordinates*/

  TO = lattice->TO;

  p = diff->p;
  l = lattice->FROM[x][y][z].count[iSpecies];

  /*  n0 = rd_round((double)gasdevT(&seed, l*p[0], l * p[0] * ( 1.0 -p[0])));*/

  n_m = mobile_particles;
  n0 = l - n_m;


  /*make three groups: one for each dimension*/
  g1 = rd_round ((double) gasdev(&seed, n_m/3.0, n_m*2.0/9.0));
  g2 = rd_round((double) gasdev(&seed, (n_m - g1)/2.0, (n_m-g1)/4.0));
  g3 = n_m - g1 - g2;

  assert(g1 >=0 && g2 >= 0);

  /*  fprintf(stderr, "%d %d\n", g1, g2);*/

  gg1 = rd_round((double) gasdev(&seed, g1 / 2.0, g1 / 4.0));
  gg12 = g1 - gg1;

  gg2 = rd_round((double) gasdev(&seed, g2 / 2.0, g2 / 4.0));
  gg22 = g2 - gg2;

  gg3 = rd_round((double) gasdev(&seed, g3 / 2.0, g3 / 4.0));
  gg32 = g3 - gg3;


  
#ifdef DEBUG_DIFF
  fprintf(stderr, "n[0]=%d, n[1]=%d, n[2]=%d, n[3]=%d, n[4]=%d, n[5]=%d, n[6]=%d\n", n0, gg1, gg12, gg2, gg22, g3, gg32);
#endif

  getNeighbourCoord(-1, x,y,z, &neigh, lattice->boundary, lattice, N);
  
  TO[x][y][z].count[iSpecies] += n0;
  /*  left = (x==0)?LSIZE-1: x-1;*/
  TO[N[0].x][N[0].y][N[0].z].count[iSpecies]+= gg1;
  /*  top = (y==0)?LSIZE-1: y-1;*/
  TO[N[1].x][N[1].y][N[1].z].count[iSpecies]+= gg2;
  /*  right = (x==LSIZE-1)?0: x+1;*/
  TO[N[2].x][N[2].y][N[2].z].count[iSpecies]+= gg12;
  
  /*  bottom = (y==LSIZE-1)?0: y+1;*/
  TO[N[3].x][N[3].y][N[3].z].count[iSpecies]+= gg22;
  
  /* front */
  TO[N[4].x][N[4].y][N[4].z].count[iSpecies]+= gg3;
  
  /* back */
  TO[N[5].x][N[5].y][N[5].z].count[iSpecies]+= gg32;
}

/* void diffusion_few(int max_l, TDiffusion *diff, int x, int y, int z, TLattice *lattice, int iSpecies){ */
/*   double r; */
/*   int j,l; */
/*   double *p; */
/*   Dimension N[6]; */
/*   int lo, hi, mid; */

/*   p = diff->p; */

/*   getNeighbourCoord(0,x,y,z, NULL, lattice->boundary, lattice, N); */
  
/*   for(l=0; l < max_l; l++){ */
/*     r = rand3(&seed);   */
/*     lo=0; hi = 2*DIMENSIONALITY; */
    
/*     while( lo <= hi ){ */
/*       mid= (lo+hi)/2; */
/*       if( r < p[mid]) */
/* 	hi = mid-1; */
/*       else if (p[mid] < r) */
/* 	lo = mid +1; */
/*     } */
/*      */
/*     diffusion_small(mid, x,y,z, lattice, iSpecies, N); */
/*   } */
/* } */

void diffusion_few(int max_l, TDiffusion *diff, int x, int y, int z, TLattice *lattice, int iSpecies){
  double r;
  int j,l;
  double *p;
  Dimension N[6];

  p = diff->p;

  getNeighbourCoord(0,x,y,z, NULL, lattice->boundary, lattice, N);

  for(l=0; l < max_l; l++){
    r = rand3(&seed);
    if( r > p[0]){
      for(j=1; j < NEIGH; j++){
	if( p[j-1] < r && r <= p[j] ){
	  diffusion_small(j, x,y,z, lattice, iSpecies, N);
	  break;
	}
      }
    } else {
      diffusion_small(0, x,y,z, lattice,iSpecies, N);  
    }
  }
}

void diffusion_small(int i, int x, int y, int z, TLattice *lattice, int iSpecies, Dimension *N){
  tSite (*TO)[L_MAX_Y_SIZE][L_MAX_Z_SIZE];
  
  TO = lattice->TO;

  switch(i){
  case 0:
    /*A current, B next*/
    TO[x][y][z].count[iSpecies]++;
    break;
  case 1: /*left = (x==0)?LSIZE-1: x-1;*/
    /*A current, B next*/
    TO[N[0].x][N[0].y][N[0].z].count[iSpecies]++;
    break;
    
  case 2: /*top = (y==0)?LSIZE-1: y-1;*/
    TO[N[1].x][N[1].y][N[1].z].count[iSpecies]++;
    break;
    
  case 3: /*right = (x==LSIZE-1)?0: x+1;*/
    TO[N[2].x][N[2].y][N[2].z].count[iSpecies]++;
    break;
    
  case 4: /*bottom = (y==LSIZE-1)?0: y+1;*/
    TO[N[3].x][N[3].y][N[3].z].count[iSpecies]++;
    break;
  case 5:
    /*front*/
    TO[N[4].x][N[4].y][N[4].z].count[iSpecies]++;
    break;
  case 6:
    /*back*/
    TO[N[5].x][N[5].y][N[5].z].count[iSpecies]++;
  }
}
