#include "gillespie.h"
#include "rng.h"
#include "rd_simulation.h"
#include<stdio.h>

extern long seed;

void gillespie(TGillespie *data, tSite *siteFrom, double t_delta, double lambda){
  int v;
  double sum_a;
  int (*cX)[10];
  int i, i0, i2, i4, i6, i8, n1, n3  ,n5, n7, n9;
  double r1,r2;
  double tau; /*when next reaction occurs */
  int j;
  int mu; /* selected reaction */
  double t; /*current simulation time*/
  /*  double tprint;*/
  static double a[MAXSPECIES];
  double *c;
  int *n ; /*, *prev;*/
  double rat;
  static double ratio[MAXSPECIES];
  double siteVolume;
  double volcube;
  t = 0.0;
  /*  tprint = 0.0;*/

  cX = data->cX;
  c = data->c;
  n = siteFrom->count;
  volcube = data->length*data->length*data->length;

  siteVolume = volcube * 1.0e-15 / (lambda*lambda*lambda) /*lambda in mu m, but volume in L. 1e-15*/;
#ifdef COLLISION
  if(n[cX[0][0]-1] >= 1 && n[cX[0][2]-1] >= 1)
    fprintf(stderr,"%d\n",n[cX[0][0]-1] + n[cX[0][2]-1]);
#endif

  do{
    /* calculate propensity probabilites */
    sum_a = 0.0;
    for(v=0; v < data->M; v++){
      if(cX[v][1] == -1 && cX[v][3] == -1){ /* bimolecular */
	i0 = cX[v][0]-1;
	i2 = cX[v][2]-1;
	a[v] =  c[v] * n[i0] * n[i2] / (siteVolume * AVOGADRO) * 1e6; /* reaction in mu M^-1 t^-1*/
	/*	fprintf(stderr, "%d\n", n[i2]);
	if( n[i2] > 1000)
	exit(0);*/
      }else if( cX[v][1] == -2){ /* monomolecular second order */
	i = cX[v][0] -1; /* get the specie to use */
	a[v] = c[v] * n[i] * (n[i] -1) / 2.0;
      } else if( cX[v][3] == -2){
	i = cX[v][0]-1;
	a[v] = c[v] * n[i] * (n[i] - 1) / 2.0;
      } else if(cX[v][1] == -1 && cX[v][3] == 0){ /*monomolecular first order */
	i = cX[v][0]-1;
	a[v] = c[v] * n[i];
      } else if(cX[v][1] == 0 && cX[v][3] == -1){
	i = cX[v][2]-1;
	a[v] = c[v] * n[i];
      } 

      sum_a += a[v];
    }


    if( sum_a == 0.0){
      /*      fprintf(stderr, "%e\n", t_delta - t);*/
      break;
    }
      

    /* select reaction */
     r1 = rand3(&seed); 
     r2 = rand3(&seed); 

    tau =  -log(r1)/ sum_a;

    /* If we overpass the time limit impossed by Diffusion,
     * then do not perform the reaction.
     */
    if( t + tau > t_delta )
      break;

    rat = 0.0;
    for(j=0; j<data->M; j++){
      ratio[j] = rat + (a[j]/sum_a);
      rat = ratio[j];
    }

    if( r2 >= 0.0 && r2 <= ratio[0]){
      mu = 0;
    } else {
      for(j=0; j< data->M-1; j++){
	if( r2 > ratio[j] && r2 <= ratio[j+1]){
	  mu = j+1;
	  break;
	}
      }
    }

    /*    fprintf(stderr, "MU = %d\n", mu);*/
    /*    if( mu > 1){
      printf("STOP");
    }
    */

    i0 = cX[mu][0]-1;
    i2 = cX[mu][2]-1;
    i4 = cX[mu][4]-1;
    i6 = cX[mu][6]-1;
    i8 = cX[mu][8]-1;
    n1 = cX[mu][1];
    n3 = cX[mu][3];
    n5 = cX[mu][5];
    n7 = cX[mu][7];
    n9 = cX[mu][9];

    if( i0 >= 0)
      n[i0] += n1;
    if( i2 >= 0)
      n[i2] += n3;
    if( i4 >= 0)
      n[i4] += n5;
    if( i6 >= 0)
      n[i6] += n7;
    if( i8 >= 0)
      n[i8] += n9;
     
      
    t += tau;

    if(data->output->vis_method == 4){ /*print time of reaction*/
      printf( "%.10lf ", data->simTime + t);
      for(i=0; i<data->N; i++){
	printf( "%d ",n[i]);
	       }
      printf("\n");
      }

  } while(t <= t_delta);

}

/* void write_results(double time, int *nMolecules, int N){ */
/*   int i; */
/*   double inf = 1.0/0.0; */
/*   if( time == inf) */
/*     return; */

/*   printf("%f", time); */

/*   for(i=0; i<N; i++) */
/*     printf(" %d", nMolecules[i]); */

/*   printf("\n"); */
/* } */


