/*
 * Multi species diffusion in a square lattice LSIZE x LSIZE
 * with refractant boundaries. Supports interface with matlab
 *
 * Using algorithms by Bastien Chopard
 * Mor efficient particle diffusion using Gaussian random numbers.
 *
 * Usage
 *   diffusion inputfile vismethod={0,1,2,3,4} print_method={0,1,2} statistics={2}
 *
 * Input data
 *
 *
 * Output data
 *  time #
 *    y  n_particles
 *
 * version 13 31.05.2006
 * - Added output at every reaction event
 *
 * version 12 9.9.2005(rdm5):
 * - Added input parameter for system size. (before it was assumed to be 1 \mu m)
 *
 * version 10 (rdm4):
 * -Added curved membrane diffusion, although the coefficient D is not valid!!!
 *
 * Version 9:
 * - Adding #rd dimension
 *
 * Verison 8:
 * - Membrane diffusion
 * - configurable lattice size
 *
 * Version 7:
 * - Diffusion uses now real clock, synchronised with reaction clock
 * - reaction only occurs once
 *
 * Version 6:
 *  - Adding simple squared membrane surrounding the whole system
 *
 * Version 5:
 *  - Split files, and structures to eaase reading.
 *  - Added Matlab interface with mx (rd_matlab.c)
 *
 * Version 4: 
 *  - Added multiple experiments and data collector information
 *  - Added mean, std error of mean analysis for mass output.
 *
 * created: 30/03/2004 by Jordi Vidal
 *
 */

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<math.h>
#include"types.h"
#include"gasdev.h"
#include"rng.h"
#include"gillespie.h"
#include"read_configuration.h"
#include"statistics.h"
#include"rd_simulation.h"

long seed = -51;


void cleanUp(TGillespie *gil){
  free(gil->c);
  free(gil->cX);
}

int main(int argc, char *argv[]){
  int experiment = 0;
  int nExperiments = 0;

  TGillespie gillespie;
  static TLattice lattice; 
  TInitConditions init; 
  TOutput output;
  TDiffusion diffusion; 
  TTimes times; 
  TGeometry geometry; 

  fprintf(stderr, "\n\n\tStochastic Reaction-Diffusion Simulator\n\t\t by Jordi Vidal, (c) 2003-2004\n\n");

  if( argc != NUM_ARGUMENTS){
    fprintf(stderr, MSG_USAGE);
    exit(-1);
  }if( argc == NUM_ARGUMENTS){
    output.vis_method = atoi(argv[2]);
    output.print_method = atoi(argv[3]);
    output.stat_method = atoi(argv[4]);
    nExperiments = atoi(argv[5]);
    gillespie.output = &output;
  }

  fprintf(stderr, "Initialization");

  read_configuration(argv[1], &gillespie, &lattice, &init, &output, &diffusion, &times, &geometry);
  times.tau_print = times.t_print;
  fprintf(stderr, ".");

  /*  output.statistics.statMass = initialize_st_mass(times.t_max, times.tau_print, gillespie.N);*/
  fprintf(stderr, ".\n");

  fprintf(stderr, "Starting experiments...\n");

  for(experiment=0; experiment < nExperiments; experiment++ ){
    fprintf(stderr, "Experiment n=%d",  experiment);

    rd_simulation(&lattice, &init, &gillespie, &output, &diffusion, &times, &geometry, experiment);

  } /*end Experiments*/

      /*  Write Statistical results collected through simulations*/
  switch(output.stat_method){
  case 0: break;
  case 1: break;
  case 2: 
    /*write_st_mass(output.statistics.statMass, times.t_max, times.tau_print, gillespie.N);*/
  default:;
  }

  cleanUp(&gillespie);
   return 1;
 }


