#ifndef __TYPES__
#define __TYPES__

#define NUM_ARGUMENTS 6
#define MSG_USAGE "Reaction-diffusion ussage:\n\
\treacdiff input_file visualization_method print_method statistics num.Experiments\n\n\
visualization method = 0 linear time increment for printing\n\
                       1 exponential time increment for printing\n\
                       2 print at t=0 and t=t_simulation time\n\
                       3 print at certain number of particles\n\
                       4 print at reaction event\n\
print_method = 0 column profile (averaged over rows) per species\n\
               1 particles count per site per species\n\
               2 mass of system per species\n\
               3 2d surface\n\
               4 radius of circle domain\n\
               5 z plane in mathematica table form\n\
statistics = 2 mass average and standard error of mean\n\n\
nExperiments = n number of experiments for averaging\n"


#define L_MAX_X_SIZE 20
#define L_MAX_Y_SIZE 20
#define L_MAX_Z_SIZE 20
#define NEIGH 7  /* 3D*/
#define DIMENSIONALITY 3
#define MAXSPECIES 3

#define FILE_VERSION 12

#define TRUE 1
#define FALSE 0
#define AVOGADRO 6.02215e+23

/* #define SS 0 */
/* #define R1 1 */
/* #define R2 2 */

/* typedef int eBoundary; */

enum eboundary {SS = 0, R1=1, R2=2, GFILE=3, QUARTER=4};
enum einitialization {HOMOGENEOUS=0, RANDOM=1, SOURCESINK=2};
typedef enum  eboundary eBoundary;
typedef enum einitialization eInitialization;

typedef short unsigned int Tid;
#define MEMBRANE 1
#define CYTOSOL 0
#define EXTERIOR 2

#define TOP 0
#define RIGHT 1
#define BOTTOM 2
#define LEFT 3
#define FRONT 4
#define BACK 5

typedef struct tsite{
  int count[MAXSPECIES];
} tSite;

typedef struct dimension{
  int x;
  int y;
  int z;
} Dimension;

typedef struct neighbours{
  int top;
  int bottom;
  int left;
  int right;
  int front;
  int back;
} iNeighbours;



typedef struct tlattice{
  tSite A[L_MAX_X_SIZE][L_MAX_Y_SIZE][L_MAX_Z_SIZE];
  tSite B[L_MAX_X_SIZE][L_MAX_Y_SIZE][L_MAX_Z_SIZE];
  tSite (*TO)[L_MAX_Y_SIZE][L_MAX_Z_SIZE];
  tSite (*FROM)[L_MAX_Y_SIZE][L_MAX_Z_SIZE];
  Tid id[L_MAX_X_SIZE][L_MAX_Y_SIZE][L_MAX_Z_SIZE];
  Dimension N[L_MAX_X_SIZE][L_MAX_Y_SIZE][L_MAX_Z_SIZE][9]; /*membrane neighbours coordinates 3D spherical*/
  int s;
  eBoundary boundary;
  int xdim;
  int ydim;
  int zdim;
} TLattice;

enum epartconc {CONCENTRATION=0, MOLECULES=1};
typedef enum epartconc ePartConc;

typedef struct tinitpart{
  ePartConc type;
  double value;
} TInitPart;

typedef struct tinitCond{
  TInitPart initCond[MAXSPECIES];
  eInitialization type; 
} TInitConditions;

typedef struct tstatistics{
  double ***statMass;
  double **mstatMass;
} TStatistics;

typedef struct toutput{
  int vis_method;
  int stat_method;
  int print_method;
  TStatistics statistics;
} TOutput;

typedef struct tdiffusion{
  double D[MAXSPECIES];
  double p[NEIGH];
  double lambda;
} TDiffusion;

typedef struct ttimes{
  double t_max;/*maximum simulation time*/
  double t_print;/*next print alarm*/
  double tau_print;/*print interval time*/
  double t_alarm[MAXSPECIES][2];/*Diffusion alarm times*/ /* [][0] absolute time for alarm, [][1] delta for alarm */
  double t_tau; /*simulation interval between diffusion*/
} TTimes;



typedef struct tgeometry{
  char filename[100];
  int cellTypeCount[3]; /*number of cytosol and membrane cells in the geometry*/
  Dimension centre;
} TGeometry;



#endif
